from collections.abc import Iterator, Sequence
from pathlib import Path
from typing import Annotated, Callable, Optional, Union

from dagster._core.definitions.assets import AssetsDefinition
from dagster._core.definitions.definitions_class import Definitions
from dagster._core.definitions.events import AssetMaterialization
from dagster._core.definitions.result import MaterializeResult
from dagster._record import record
from dagster_sling import DagsterSlingTranslator, SlingResource, sling_assets
from dagster_sling.resources import AssetExecutionContext
from typing_extensions import Self

from dagster_components import Component, ComponentLoadContext
from dagster_components.core.component import registered_component_type
from dagster_components.core.component_scaffolder import ComponentScaffolder
from dagster_components.core.schema.metadata import ResolvableFieldInfo
from dagster_components.core.schema.objects import (
    AssetAttributesModel,
    AssetSpecTransformModel,
    OpSpecModel,
    ResolvableModel,
)
from dagster_components.core.schema.resolver import TemplatedValueResolver
from dagster_components.utils import ResolvingInfo, get_wrapped_translator_class


@record
class SlingReplicationSpec:
    relative_path: str
    op_spec: OpSpecModel
    translator: DagsterSlingTranslator


class SlingReplicationParams(ResolvableModel[SlingReplicationSpec]):
    path: str
    op: Optional[OpSpecModel] = None
    asset_attributes: Annotated[
        Optional[AssetAttributesModel], ResolvableFieldInfo(required_scope={"stream_definition"})
    ] = None

    def resolve(self, resolver: TemplatedValueResolver) -> SlingReplicationSpec:
        return SlingReplicationSpec(
            relative_path=resolver.resolve_obj(self.path),
            op_spec=self.op.resolve(resolver) if self.op else OpSpecModel(),
            translator=get_wrapped_translator_class(DagsterSlingTranslator)(
                resolving_info=ResolvingInfo(
                    "stream_definition", self.asset_attributes or AssetAttributesModel(), resolver
                ),
            ),
        )


ResolvedSlingReplicationCollectionParams = tuple[
    SlingResource,
    Sequence[SlingReplicationSpec],
    Sequence[Callable[[Definitions], Definitions]],
]


class SlingReplicationCollectionParams(ResolvableModel[ResolvedSlingReplicationCollectionParams]):
    sling: Optional[SlingResource] = None
    replications: Sequence[SlingReplicationParams]
    transforms: Optional[Sequence[AssetSpecTransformModel]] = None

    def resolve(self, resolver: TemplatedValueResolver) -> ResolvedSlingReplicationCollectionParams:
        return (
            self.sling if self.sling else SlingResource(),
            [replication.resolve(resolver) for replication in self.replications],
            [transform.resolve(resolver) for transform in self.transforms]
            if self.transforms
            else [],
        )


@registered_component_type
class SlingReplicationCollection(Component):
    """Expose one or more Sling replications to Dagster as assets."""

    def __init__(
        self,
        dirpath: Path,
        resource: SlingResource,
        replication_specs: Sequence[SlingReplicationSpec],
        transforms: Sequence[Callable[[Definitions], Definitions]],
    ):
        self.dirpath = dirpath
        self.resource = resource
        self.replication_specs = replication_specs
        self.transforms = transforms

    @classmethod
    def get_scaffolder(cls) -> ComponentScaffolder:
        from dagster_components.lib.sling_replication_collection.scaffolder import (
            SlingReplicationComponentScaffolder,
        )

        return SlingReplicationComponentScaffolder()

    @classmethod
    def get_schema(cls) -> type[SlingReplicationCollectionParams]:
        return SlingReplicationCollectionParams

    @classmethod
    def load(cls, params: SlingReplicationCollectionParams, context: ComponentLoadContext) -> Self:
        resource, replication_specs, transforms = params.resolve(context.templated_value_resolver)
        return cls(
            dirpath=context.path,
            resource=resource,
            replication_specs=replication_specs,
            transforms=transforms,
        )

    def build_asset(
        self, context: ComponentLoadContext, replication_spec: SlingReplicationSpec
    ) -> AssetsDefinition:
        @sling_assets(
            name=replication_spec.op_spec.name or Path(replication_spec.relative_path).stem,
            op_tags=replication_spec.op_spec.tags,
            replication_config=self.dirpath / replication_spec.relative_path,
            dagster_sling_translator=replication_spec.translator,
        )
        def _asset(context: AssetExecutionContext):
            yield from self.execute(context=context, sling=self.resource)

        return _asset

    def execute(
        self, context: AssetExecutionContext, sling: SlingResource
    ) -> Iterator[Union[AssetMaterialization, MaterializeResult]]:
        yield from sling.replicate(context=context)

    def build_defs(self, context: ComponentLoadContext) -> Definitions:
        defs = Definitions(
            assets=[
                self.build_asset(context, replication) for replication in self.replication_specs
            ],
        )
        for transform in self.transforms:
            defs = transform(defs)
        return defs



from pathlib import Path
from typing import Optional

from dagster._core.definitions.definitions_class import Definitions
from dagster._core.definitions.module_loaders.load_defs_from_module import (
    load_definitions_from_module,
)
from dagster._seven import import_uncached_module_from_path
from dagster._utils import pushd
from pydantic import BaseModel
from typing_extensions import Self

from dagster_components import Component, ComponentLoadContext, registered_component_type
from dagster_components.lib.definitions_component.scaffolder import DefinitionsComponentScaffolder


class DefinitionsParamSchema(BaseModel):
    definitions_path: Optional[str] = None


@registered_component_type(name="definitions")
class DefinitionsComponent(Component):
    """Wraps an arbitrary set of Dagster definitions."""

    def __init__(self, definitions_path: Path):
        self.definitions_path = definitions_path

    @classmethod
    def get_scaffolder(cls) -> DefinitionsComponentScaffolder:
        return DefinitionsComponentScaffolder()

    @classmethod
    def get_schema(cls) -> type[DefinitionsParamSchema]:
        return DefinitionsParamSchema

    @classmethod
    def load(cls, params: DefinitionsParamSchema, context: ComponentLoadContext) -> Self:
        return cls(definitions_path=Path(params.definitions_path or "definitions.py"))

    def build_defs(self, context: ComponentLoadContext) -> Definitions:
        with pushd(str(context.path)):
            module = import_uncached_module_from_path("definitions", str(self.definitions_path))

        return load_definitions_from_module(module)



"""Sample local components for testing validation. Paired with test cases
in integration_tests/components/validation.
"""

from dagster._core.definitions.definitions_class import Definitions
from pydantic import BaseModel, ConfigDict
from typing_extensions import Self

from dagster_components import Component, registered_component_type
from dagster_components.core.component import ComponentLoadContext


class MyComponentSchema(BaseModel):
    a_string: str
    an_int: int

    model_config = ConfigDict(extra="forbid")


@registered_component_type
class MyComponent(Component):
    name = "my_component"

    @classmethod
    def get_schema(cls) -> type[MyComponentSchema]:
        return MyComponentSchema

    @classmethod
    def load(cls, params: MyComponentSchema, context: ComponentLoadContext) -> Self:
        return cls()

    def build_defs(self, context: ComponentLoadContext) -> Definitions:
        return Definitions()


class MyNestedModel(BaseModel):
    a_string: str
    an_int: int

    model_config = ConfigDict(extra="forbid")


class MyNestedComponentSchema(BaseModel):
    nested: dict[str, MyNestedModel]

    model_config = ConfigDict(extra="forbid")


@registered_component_type
class MyNestedComponent(Component):
    name = "my_nested_component"

    @classmethod
    def get_schema(cls) -> type[MyNestedComponentSchema]:
        return MyNestedComponentSchema

    @classmethod
    def load(cls, params: MyComponentSchema, context: ComponentLoadContext) -> Self:
        return cls()

    def build_defs(self, context: ComponentLoadContext) -> Definitions:
        return Definitions()



from collections.abc import Mapping, Sequence
from typing import Annotated, Any, Callable, Literal, Optional, Union

import dagster._check as check
from dagster._core.definitions.asset_key import AssetKey
from dagster._core.definitions.asset_selection import AssetSelection
from dagster._core.definitions.asset_spec import AssetSpec, map_asset_specs
from dagster._core.definitions.assets import AssetsDefinition
from dagster._core.definitions.declarative_automation.automation_condition import (
    AutomationCondition,
)
from dagster._core.definitions.definitions_class import Definitions
from dagster._record import replace
from pydantic import BaseModel

from dagster_components.core.schema.base import ResolvableModel
from dagster_components.core.schema.metadata import ResolvableFieldInfo
from dagster_components.core.schema.resolver import TemplatedValueResolver


class OpSpecModel(ResolvableModel["OpSpecModel"]):
    name: Optional[str] = None
    tags: Optional[dict[str, str]] = None

    def resolve(self, resolver: TemplatedValueResolver) -> "OpSpecModel":
        return OpSpecModel(**resolver.resolve_obj(self.model_dump(exclude_unset=True)))


class _ResolvableAssetAttributesMixin(BaseModel):
    deps: Sequence[str] = []
    description: Optional[str] = None
    metadata: Annotated[
        Union[str, Mapping[str, Any]], ResolvableFieldInfo(output_type=Mapping[str, Any])
    ] = {}
    group_name: Optional[str] = None
    skippable: bool = False
    code_version: Optional[str] = None
    owners: Sequence[str] = []
    tags: Annotated[
        Union[str, Mapping[str, str]], ResolvableFieldInfo(output_type=Mapping[str, str])
    ] = {}
    kinds: Optional[Sequence[str]] = None
    automation_condition: Annotated[
        Optional[str], ResolvableFieldInfo(output_type=Optional[AutomationCondition])
    ] = None


class AssetAttributesModel(_ResolvableAssetAttributesMixin, ResolvableModel[Mapping[str, Any]]):
    key: Annotated[Optional[str], ResolvableFieldInfo(output_type=AssetKey)] = None

    def resolve(self, resolver: TemplatedValueResolver) -> Mapping[str, Any]:
        props = resolver.resolve_obj(self.model_dump(exclude_unset=True))
        if "key" in props:
            props["key"] = AssetKey.from_user_string(props["key"])
        return props


class AssetSpecModel(_ResolvableAssetAttributesMixin, ResolvableModel[AssetSpec]):
    key: Annotated[Optional[str], ResolvableFieldInfo(output_type=AssetKey)]

    def resolve(self, resolver: TemplatedValueResolver) -> AssetSpec:
        return AssetSpec(
            key=AssetKey.from_user_string(resolver.resolve_obj(self.key)),
            description=resolver.resolve_obj(self.description),
            metadata=resolver.resolve_obj(self.metadata),
            group_name=resolver.resolve_obj(self.group_name),
            skippable=resolver.resolve_obj(self.skippable),
            code_version=resolver.resolve_obj(self.code_version),
            owners=resolver.resolve_obj(self.owners),
            tags=resolver.resolve_obj(self.tags),
            deps=[AssetKey.from_user_string(d) for d in resolver.resolve_obj(self.deps)]
            if self.deps
            else None,
        )


class AssetSpecTransformModel(ResolvableModel[Callable[[Definitions], Definitions]]):
    target: str = "*"
    operation: Literal["merge", "replace"] = "merge"
    attributes: AssetAttributesModel

    class Config:
        arbitrary_types_allowed = True

    def apply_to_spec(
        self,
        spec: AssetSpec,
        value_resolver: TemplatedValueResolver,
    ) -> AssetSpec:
        # add the original spec to the context and resolve values
        attributes = self.attributes.resolve(value_resolver.with_scope(asset=spec))

        if self.operation == "merge":
            mergeable_attributes = {"metadata", "tags"}
            merge_attributes = {k: v for k, v in attributes.items() if k in mergeable_attributes}
            replace_attributes = {
                k: v for k, v in attributes.items() if k not in mergeable_attributes
            }
            return spec.merge_attributes(**merge_attributes).replace_attributes(
                **replace_attributes
            )
        elif self.operation == "replace":
            return spec.replace_attributes(**attributes)
        else:
            check.failed(f"Unsupported operation: {self.operation}")

    def apply(self, defs: Definitions, value_resolver: TemplatedValueResolver) -> Definitions:
        target_selection = AssetSelection.from_string(self.target, include_sources=True)
        target_keys = target_selection.resolve(defs.get_asset_graph())

        mappable = [d for d in defs.assets or [] if isinstance(d, (AssetsDefinition, AssetSpec))]
        mapped_assets = map_asset_specs(
            lambda spec: self.apply_to_spec(spec, value_resolver)
            if spec.key in target_keys
            else spec,
            mappable,
        )

        assets = [
            *mapped_assets,
            *[d for d in defs.assets or [] if not isinstance(d, (AssetsDefinition, AssetSpec))],
        ]
        return replace(defs, assets=assets)

    def resolve(self, resolver: TemplatedValueResolver) -> Callable[[Definitions], Definitions]:
        return lambda defs: self.apply(defs, resolver)



import inspect
from collections.abc import Mapping, Sequence
from pathlib import Path
from typing import Any, Optional, TypeVar, Union

from dagster._core.errors import DagsterInvalidDefinitionError
from dagster._record import record
from dagster._utils import pushd
from dagster._utils.pydantic_yaml import (
    _parse_and_populate_model_with_annotated_errors,
    enrich_validation_errors_with_source_position,
)
from dagster._utils.source_position import SourcePositionTree
from dagster._utils.yaml_utils import parse_yaml_with_source_positions
from pydantic import BaseModel, TypeAdapter

from dagster_components.core.component import (
    Component,
    ComponentDeclNode,
    ComponentLoadContext,
    ComponentTypeRegistry,
    get_component_type_name,
    is_component_loader,
    is_registered_component_type,
)
from dagster_components.utils import load_module_from_path


class ComponentFileModel(BaseModel):
    type: str
    params: Optional[Mapping[str, Any]] = None


T = TypeVar("T", bound=BaseModel)


def find_local_component_types(component_path: Path) -> list[type[Component]]:
    """Find all component types defined in a component directory."""
    component_types = []
    for py_file in component_path.glob("*.py"):
        module_name = py_file.stem

        module = load_module_from_path(module_name, component_path / f"{module_name}.py")

        for _name, obj in inspect.getmembers(module, inspect.isclass):
            assert isinstance(obj, type)
            if is_registered_component_type(obj):
                component_types.append(obj)
    return component_types


@record
class PythonComponentDecl(ComponentDeclNode):
    path: Path

    @staticmethod
    def component_file_path(path: Path) -> Path:
        return path / "component.py"

    @staticmethod
    def exists_at(path: Path) -> bool:
        return PythonComponentDecl.component_file_path(path).exists()

    @staticmethod
    def from_path(path: Path) -> "PythonComponentDecl":
        return PythonComponentDecl(path=path)

    def load(self, context: ComponentLoadContext) -> Sequence[Component]:
        module = load_module_from_path(
            self.path.stem, PythonComponentDecl.component_file_path(self.path)
        )
        component_loaders = list(inspect.getmembers(module, is_component_loader))
        if len(component_loaders) < 1:
            raise DagsterInvalidDefinitionError("No component loaders found in module")
        elif len(component_loaders) > 1:
            # note: we could support multiple component loaders in the same file, just
            # being more restrictive to start
            raise DagsterInvalidDefinitionError(
                f"Multiple component loaders found in module: {component_loaders}"
            )
        else:
            _, component_loader = component_loaders[0]
            return [component_loader(context)]


@record
class YamlComponentDecl(ComponentDeclNode):
    path: Path
    component_file_model: ComponentFileModel
    source_position_tree: Optional[SourcePositionTree] = None

    @staticmethod
    def component_file_path(path: Path) -> Path:
        return path / "component.yaml"

    @staticmethod
    def exists_at(path: Path) -> bool:
        return YamlComponentDecl.component_file_path(path).exists()

    @staticmethod
    def from_path(path: Path) -> "YamlComponentDecl":
        component_file_path = YamlComponentDecl.component_file_path(path)
        parsed = parse_yaml_with_source_positions(
            component_file_path.read_text(), str(component_file_path)
        )
        obj = _parse_and_populate_model_with_annotated_errors(
            cls=ComponentFileModel, obj_parse_root=parsed, obj_key_path_prefix=[]
        )

        return YamlComponentDecl(
            path=path,
            component_file_model=obj,
            source_position_tree=parsed.source_position_tree,
        )

    def get_component_type(self, registry: ComponentTypeRegistry) -> type[Component]:
        parsed_defs = self.component_file_model
        if parsed_defs.type.startswith("."):
            component_registry_key = parsed_defs.type[1:]

            for component_type in find_local_component_types(self.path):
                if get_component_type_name(component_type) == component_registry_key:
                    return component_type

            raise Exception(
                f"Could not find component type {component_registry_key} in {self.path}"
            )

        return registry.get(parsed_defs.type)

    def get_params(self, context: ComponentLoadContext, params_schema: type[T]) -> T:
        with pushd(str(self.path)):
            if self.source_position_tree:
                source_position_tree_of_params = self.source_position_tree.children["params"]
                with enrich_validation_errors_with_source_position(
                    source_position_tree_of_params, ["params"]
                ):
                    return TypeAdapter(params_schema).validate_python(
                        self.component_file_model.params
                    )
            else:
                return TypeAdapter(params_schema).validate_python(self.component_file_model.params)

    def load(self, context: ComponentLoadContext) -> Sequence[Component]:
        component_type = self.get_component_type(context.registry)
        component_schema = component_type.get_schema()
        context = context.with_rendering_scope(component_type.get_additional_scope())
        loaded_params = self.get_params(context, component_schema) if component_schema else None
        return [component_type.load(loaded_params, context)]


@record
class ComponentFolder(ComponentDeclNode):
    path: Path
    sub_decls: Sequence[Union[YamlComponentDecl, PythonComponentDecl, "ComponentFolder"]]

    def load(self, context: ComponentLoadContext) -> Sequence[Component]:
        components = []
        for sub_decl in self.sub_decls:
            sub_context = context.for_decl_node(sub_decl)
            components.extend(sub_decl.load(sub_context))
        return components


def path_to_decl_node(path: Path) -> Optional[ComponentDeclNode]:
    # right now, we only support two types of components, both of which are folders
    # if the folder contains a component.yaml file, it's a component instance
    # otherwise, it's a folder containing sub-components

    if not path.is_dir():
        return None

    if YamlComponentDecl.exists_at(path):
        return YamlComponentDecl.from_path(path)
    elif PythonComponentDecl.exists_at(path):
        return PythonComponentDecl.from_path(path)

    subs = []
    for subpath in path.iterdir():
        component = path_to_decl_node(subpath)
        if component:
            subs.append(component)

    return ComponentFolder(path=path, sub_decls=subs)



from abc import abstractmethod
from collections.abc import Mapping
from typing import Any, Generic, TypeVar

from pydantic import BaseModel, ConfigDict, TypeAdapter

from dagster_components.core.schema.metadata import get_resolution_metadata
from dagster_components.core.schema.resolver import TemplatedValueResolver

T = TypeVar("T")


class ResolvableModel(BaseModel, Generic[T]):
    """Base class for models that are part of a component schema."""

    model_config = ConfigDict(extra="forbid")

    def _resolved_properties(self, value_resolver: TemplatedValueResolver) -> Mapping[str, Any]:
        """Returns a dictionary of resolved properties for this class."""
        raw_properties = self.model_dump(exclude_unset=True)

        # validate that the resolved properties match the output type
        resolved_properties = {}
        for k, v in raw_properties.items():
            resolved = value_resolver.resolve_obj(v)
            annotation = self.__annotations__[k]
            rendering_metadata = get_resolution_metadata(annotation)

            if rendering_metadata.post_process:
                resolved = rendering_metadata.post_process(resolved)

            # hook into pydantic's type validation to handle complicated stuff like Optional[Mapping[str, int]]
            TypeAdapter(
                rendering_metadata.output_type, config={"arbitrary_types_allowed": True}
            ).validate_python(resolved)

            resolved_properties[k] = resolved

        return resolved_properties

    @abstractmethod
    def resolve(self, value_resolver: TemplatedValueResolver) -> T: ...



import os
from pathlib import Path
from typing import Optional

import dagster._check as check
from dbt.cli.main import dbtRunner
from pydantic import BaseModel, Field

from dagster_components.core.component_scaffolder import (
    ComponentScaffolder,
    ComponentScaffoldRequest,
)
from dagster_components.scaffold import scaffold_component_yaml


class DbtScaffoldParams(BaseModel):
    init: bool = Field(default=False)
    project_path: Optional[str] = None


class DbtProjectComponentScaffolder(ComponentScaffolder):
    @classmethod
    def get_schema(cls) -> Optional[type[BaseModel]]:
        return DbtScaffoldParams

    def scaffold(self, request: ComponentScaffoldRequest, params: DbtScaffoldParams) -> None:
        cwd = os.getcwd()
        if params.project_path:
            # NOTE: CWD is not set "correctly" above so we prepend "../../.." as a temporary hack to
            # make sure the path is right.
            relative_path = os.path.join(
                "../../../", os.path.relpath(params.project_path, start=cwd)
            )
        elif params.init:
            dbtRunner().invoke(["init"])
            subpaths = [
                path for path in Path(cwd).iterdir() if path.is_dir() and path.name != "logs"
            ]
            check.invariant(len(subpaths) == 1, "Expected exactly one subpath to be created.")
            # this path should be relative to this directory
            relative_path = subpaths[0].name
        else:
            relative_path = None

        scaffold_component_yaml(request, {"dbt": {"project_dir": relative_path}})



import importlib.util
import sys
from collections.abc import Iterator, Mapping
from contextlib import contextmanager
from dataclasses import dataclass
from pathlib import Path
from types import ModuleType
from typing import Any, Optional

import click
from dagster._core.definitions.asset_key import AssetKey
from dagster._core.definitions.asset_spec import AssetSpec
from dagster._core.definitions.declarative_automation.automation_condition import (
    AutomationCondition,
)
from dagster._core.errors import DagsterError

from dagster_components.core.schema.objects import AssetAttributesModel
from dagster_components.core.schema.resolver import TemplatedValueResolver

CLI_BUILTIN_COMPONENT_LIB_KEY = "builtin_component_lib"


def ensure_dagster_components_tests_import() -> None:
    from dagster_components import __file__ as dagster_components_init_py

    dagster_components_package_root = (Path(dagster_components_init_py) / ".." / "..").resolve()
    assert (
        dagster_components_package_root / "dagster_components_tests"
    ).exists(), "Could not find dagster_components_tests where expected"
    sys.path.append(dagster_components_package_root.as_posix())


def exit_with_error(error_message: str) -> None:
    click.echo(click.style(error_message, fg="red"))
    sys.exit(1)


# Temporarily places a path at the front of sys.path, ensuring that any modules in that path are
# importable.
@contextmanager
def ensure_loadable_path(path: Path) -> Iterator[None]:
    orig_path = sys.path.copy()
    sys.path.insert(0, str(path))
    try:
        yield
    finally:
        sys.path = orig_path


def get_path_for_package(package_name: str) -> str:
    spec = importlib.util.find_spec(package_name)
    if not spec:
        raise DagsterError(f"Cannot find package: {package_name}")
    # file_path = spec.origin
    submodule_search_locations = spec.submodule_search_locations
    if not submodule_search_locations:
        raise DagsterError(f"Package does not have any locations for submodules: {package_name}")
    return submodule_search_locations[0]


@dataclass
class ResolvingInfo:
    obj_name: str
    asset_attributes: AssetAttributesModel
    value_resolver: TemplatedValueResolver

    def get_resolved_attribute(self, attribute: str, obj: Any, default_method) -> Any:
        renderer = self.value_resolver.with_scope(**{self.obj_name: obj})
        rendered_attributes = self.asset_attributes.resolve(renderer)
        return (
            rendered_attributes[attribute]
            if attribute in rendered_attributes
            else default_method(obj)
        )

    def get_asset_spec(self, base_spec: AssetSpec, context: Mapping[str, Any]) -> AssetSpec:
        """Returns an AssetSpec that combines the base spec with attributes resolved using the provided context.

        Usage:

        ```python
        class WrappedDagsterXTranslator(DagsterXTranslator):
            def __init__(self, *, base_translator, resolving_info: ResolvingInfo):
                self.base_translator = base_translator
                self.resolving_info = resolving_info

            def get_asset_spec(self, base_spec: AssetSpec, x_params: Any) -> AssetSpec:
                return self.resolving_info.get_asset_spec(
                    base_spec, {"x_params": x_params}
                )

        ```
        """
        resolver = self.value_resolver.with_scope(**context)
        resolved_attributes = self.asset_attributes.resolve(resolver)
        return base_spec.replace_attributes(**resolved_attributes)


def get_wrapped_translator_class(translator_type: type):
    """Temporary hack to allow wrapping of many methods of a given translator class. Will be removed
    once all translators implement `get_asset_spec`.
    """

    class WrappedTranslator(translator_type):
        def __init__(self, *, resolving_info: ResolvingInfo):
            self.base_translator = translator_type()
            self.resolving_info = resolving_info

        def get_asset_key(self, obj: Any) -> AssetKey:
            return self.resolving_info.get_resolved_attribute(
                "key", obj, self.base_translator.get_asset_key
            )

        def get_group_name(self, obj: Any) -> Optional[str]:
            return self.resolving_info.get_resolved_attribute(
                "group_name", obj, self.base_translator.get_group_name
            )

        def get_tags(self, obj: Any) -> Mapping[str, str]:
            return self.resolving_info.get_resolved_attribute(
                "tags", obj, self.base_translator.get_tags
            )

        def get_automation_condition(self, obj: Any) -> Optional[AutomationCondition]:
            return self.resolving_info.get_resolved_attribute(
                "automation_condition", obj, self.base_translator.get_automation_condition
            )

    return WrappedTranslator


def load_module_from_path(module_name, path) -> ModuleType:
    # Create a spec from the file path
    spec = importlib.util.spec_from_file_location(module_name, path)
    if spec is None:
        raise ImportError(f"Cannot create a module spec from path: {path}")

    # Create and load the module
    module = importlib.util.module_from_spec(spec)
    assert spec.loader, "Must have a loader"
    spec.loader.exec_module(module)
    return module



import shutil
from collections.abc import Mapping, Sequence
from pathlib import Path
from typing import TYPE_CHECKING

from dagster._core.definitions.asset_spec import AssetSpec
from dagster._core.definitions.assets import AssetsDefinition
from dagster._core.definitions.decorators.asset_decorator import multi_asset
from dagster._core.execution.context.asset_execution_context import AssetExecutionContext
from dagster._core.pipes.subprocess import PipesSubprocessClient
from pydantic import BaseModel

from dagster_components.core.component import (
    Component,
    ComponentLoadContext,
    registered_component_type,
)
from dagster_components.core.schema.base import ResolvableModel
from dagster_components.core.schema.objects import AssetSpecModel
from dagster_components.core.schema.resolver import TemplatedValueResolver

if TYPE_CHECKING:
    from dagster._core.definitions.definitions_class import Definitions


class PipesSubprocessScriptParams(BaseModel):
    path: str
    assets: Sequence[AssetSpecModel]


class PipesSubprocessScriptCollectionParams(ResolvableModel[Mapping[Path, Sequence[AssetSpec]]]):
    scripts: Sequence[PipesSubprocessScriptParams]

    def resolve(self, resolver: TemplatedValueResolver) -> Mapping[str, Sequence[AssetSpec]]:
        return {
            resolver.resolve_obj(script.path): [
                asset_model.resolve(resolver) for asset_model in script.assets
            ]
            for script in self.scripts
        }


@registered_component_type(name="pipes_subprocess_script_collection")
class PipesSubprocessScriptCollection(Component):
    """Assets that wrap Python scripts executed with Dagster's PipesSubprocessClient."""

    def __init__(self, dirpath: Path, specs_by_path: Mapping[str, Sequence[AssetSpec]]):
        self.dirpath = dirpath
        # mapping from the script name (e.g. /path/to/script_abc.py -> script_abc)
        # to the specs it produces
        self.specs_by_path = specs_by_path

    @staticmethod
    def introspect_from_path(path: Path) -> "PipesSubprocessScriptCollection":
        path_specs = {str(path): [AssetSpec(path.stem)] for path in list(path.rglob("*.py"))}
        return PipesSubprocessScriptCollection(dirpath=path, specs_by_path=path_specs)

    @classmethod
    def get_schema(cls) -> type[PipesSubprocessScriptCollectionParams]:
        return PipesSubprocessScriptCollectionParams

    @classmethod
    def load(
        cls, params: PipesSubprocessScriptCollectionParams, context: ComponentLoadContext
    ) -> "PipesSubprocessScriptCollection":
        return cls(
            dirpath=context.path, specs_by_path=params.resolve(context.templated_value_resolver)
        )

    def build_defs(self, context: "ComponentLoadContext") -> "Definitions":
        from dagster._core.definitions.definitions_class import Definitions

        return Definitions(
            assets=[
                self._create_asset_def(self.dirpath / path, specs)
                for path, specs in self.specs_by_path.items()
            ],
        )

    def _create_asset_def(self, path: Path, specs: Sequence[AssetSpec]) -> AssetsDefinition:
        # TODO: allow name paraeterization
        @multi_asset(specs=specs, name=f"script_{path.stem}")
        def _asset(context: AssetExecutionContext):
            cmd = [shutil.which("python"), path]
            return PipesSubprocessClient().run(command=cmd, context=context).get_results()

        return _asset



import importlib.util

_has_dagster_dbt = importlib.util.find_spec("dagster_dbt") is not None
_has_dagster_sling = importlib.util.find_spec("dagster_sling") is not None

if _has_dagster_dbt:
    from dagster_components.lib.dbt_project.component import (
        DbtProjectComponent as DbtProjectComponent,
    )

if _has_dagster_sling:
    from dagster_components.lib.sling_replication_collection.component import (
        SlingReplicationCollection as SlingReplicationCollection,
    )

from dagster_components.lib.definitions_component.component import (
    DefinitionsComponent as DefinitionsComponent,
)
from dagster_components.lib.pipes_subprocess_script_collection import (
    PipesSubprocessScriptCollection as PipesSubprocessScriptCollection,
)



from collections.abc import Mapping, Sequence
from pathlib import Path
from typing import TYPE_CHECKING, Optional

from dagster._utils.warnings import suppress_dagster_warnings

from dagster_components.core.component import (
    Component,
    ComponentLoadContext,
    ComponentTypeRegistry,
    TemplatedValueResolver,
)
from dagster_components.core.component_decl_builder import (
    ComponentDeclNode,
    ComponentFolder,
    YamlComponentDecl,
    path_to_decl_node,
)

if TYPE_CHECKING:
    from dagster._core.definitions.definitions_class import Definitions


def resolve_decl_node_to_yaml_decls(decl: ComponentDeclNode) -> list[YamlComponentDecl]:
    if isinstance(decl, YamlComponentDecl):
        return [decl]
    elif isinstance(decl, ComponentFolder):
        leaf_decls = []
        for sub_decl in decl.sub_decls:
            leaf_decls.extend(resolve_decl_node_to_yaml_decls(sub_decl))
        return leaf_decls

    raise NotImplementedError(f"Unknown component type {decl}")


def build_components_from_component_folder(
    context: ComponentLoadContext, path: Path
) -> Sequence[Component]:
    component_folder = path_to_decl_node(path)
    assert isinstance(component_folder, ComponentFolder)
    return component_folder.load(context.for_decl_node(component_folder))


def build_defs_from_component_path(
    path: Path,
    registry: ComponentTypeRegistry,
    resources: Mapping[str, object],
) -> "Definitions":
    """Build a definitions object from a folder within the components hierarchy."""
    decl_node = path_to_decl_node(path=path)
    if not decl_node:
        raise Exception(f"No component found at path {path}")

    context = ComponentLoadContext(
        resources=resources,
        registry=registry,
        decl_node=decl_node,
        templated_value_resolver=TemplatedValueResolver.default(),
    )
    components = decl_node.load(context)
    return defs_from_components(resources=resources, context=context, components=components)


@suppress_dagster_warnings
def defs_from_components(
    *,
    context: ComponentLoadContext,
    components: Sequence[Component],
    resources: Mapping[str, object],
) -> "Definitions":
    from dagster._core.definitions.definitions_class import Definitions

    return Definitions.merge(
        *[
            *[
                c.build_defs(context.with_rendering_scope(c.get_additional_scope()))
                for c in components
            ],
            Definitions(resources=resources),
        ]
    )


# Public method so optional Nones are fine
@suppress_dagster_warnings
def build_component_defs(
    components_root: Path,
    resources: Optional[Mapping[str, object]] = None,
    registry: Optional["ComponentTypeRegistry"] = None,
) -> "Definitions":
    """Build a Definitions object for all the component instances in a given code location.

    Args:
        components_root (Path): The path to the components root. This is a directory containing
            subdirectories with component instances.
    """
    from dagster._core.definitions.definitions_class import Definitions

    registry = registry or ComponentTypeRegistry.from_entry_point_discovery()

    all_defs: list[Definitions] = []
    for component_path in components_root.iterdir():
        defs = build_defs_from_component_path(
            path=component_path,
            registry=registry,
            resources=resources or {},
        )
        all_defs.append(defs)
    return Definitions.merge(*all_defs)



from dagster_components.lib.test.all_metadata_empty_asset import (
    AllMetadataEmptyAsset as AllMetadataEmptyAsset,
)
from dagster_components.lib.test.complex_schema_asset import (
    ComplexSchemaAsset as ComplexSchemaAsset,
)
from dagster_components.lib.test.simple_asset import SimpleAsset as SimpleAsset
from dagster_components.lib.test.simple_pipes_script_asset import (
    SimplePipesScriptAsset as SimplePipesScriptAsset,
)
